import numpy as np
from nautilus import Sampler
from .model_galaxy import modelgalaxy
from .utils import filter_set
import time


class galaxyfitter(object):

    # Generative function for the likelihood (calls model galaxy)
    def gen_phot(self, fit_instruc, filt_files):
        """ Linear function.

        Parameters
        ----------
        
        par : numpy.ndarray
            Free parameters.
        """
        model = modelgalaxy(fit_instruc)
        wavs, flux = model.get_spectrum()
        effwavs, phot = model.get_photometry(wavs, flux, filt_files)

        return phot

    # Liklihood function for a certain point in parameter space
    def log_likelihood(self, data, gen_func, filt_files, param_dict):
        """ Return the natural logarithm of the Likelihood. Constructed from a single gaussian.

        Parameters
        ----------

        data :  numpy.ndarray
            Dataset containing x, y and y_uncert values arranged in 3 rows.

        gen_func : function
            Generative function used to model the data.

        params : numpy.ndarray
            Free parameters of the generative model.
        """

        # Given Data
        yi, yisig = data[0], data[1]

        exp = {}
        exp["age"] = param_dict["age"]
        exp["tau"] = param_dict["tau"]
        exp["massformed"] = param_dict["massformed"]
        #exp["metallicity"] = param_dict["metallicity"]        # Z/Z_oldsolar
        dust = {}
        dust["type"] = "Calzetti"
        dust["Av"] = param_dict["Av"]
        fit_instruc = {}
        fit_instruc["redshift"] = param_dict["redshift"]
        fit_instruc["exponential"] = exp   
        fit_instruc["dust"] = dust


        # Likelihood of belonging to the 'good' gaussian
        log_li = np.log((1/(np.sqrt(2*np.pi*(yisig**2))))) + ((-(yi - gen_func(fit_instruc, filt_files))**2) / (2*(yisig**2)))

        # Sum likelihood and take logarithm (note sometimes recieve underflow error due to small likelihoods)
        log_l_total = np.sum(log_li)
        
        return(log_l_total)

    # Fit an array of photometry, output the best galaxy parameters
    def fit(self, photdata, prior, filt_files):

        filt = filter_set(filt_files)
        effwavs = filt.eff_wavs

        # Run fitter
        t0 = time.time()

        # Load data
        dat= photdata.T
        dat[0] = 2.99792458E-05 * ((1e-6 * dat[0]) / ((effwavs)**2))
        dat[1] = 2.99792458E-05 * ((1e-6 * dat[1]) / ((effwavs)**2))

        # Run the nested sampler over the likelihood function
        sampler = Sampler(prior, lambda param_dict: self.log_likelihood(dat, self.gen_phot, filt_files, param_dict), n_live=1000)
        success = sampler.run(verbose=True, discard_exploration=True, timeout=1800)

        # Plot results
        points, log_w, log_l = sampler.posterior()

        dic = {}
        # Plot bestfit
        for i in range(len(prior.keys)):
            key = prior.keys[i]
            points_i = points.T[i]
            samples = np.random.choice(points_i, size = points_i.shape[0], p=np.exp(log_w))
            dic[key + "_50"] = np.percentile(samples, 50)

        exp = {}
        exp["age"] = dic['age_50']
        exp["tau"] = dic['tau_50']
        exp["massformed"] = dic['massformed_50']
        dust = {}
        dust["type"] = "Calzetti"
        dust["Av"] = dic['Av_50']
        fit_instructions = {}
        fit_instructions["redshift"] = dic['redshift_50']
        fit_instructions["exponential"] = exp   
        fit_instructions["dust"] = dust

        print('Finished in ' + str(time.time() - t0)+ ' seconds.')
        print(' ')

        return fit_instructions









import numpy as np
import matplotlib.pyplot as plt
from astropy.cosmology import Planck13  # Planck 2013
from astropy.io import fits
from spectres import spectres
from .utils import filter_set
import os


# Load SSPs and resample
def resample_ssp_wavs(target_wavs):

    grid_raw_ages = np.zeros((master_base.shape[0], master_base.shape[1], target_wavs.shape[0]+1))
    grid_raw_ages[:, :, 0] = master_base[:, :, 0]
    grid_raw_ages[:, 0, 1:] = target_wavs

    old_wavs = master_base[0][0][1:]

    for i in range(master_base.shape[0]):
        for j in range(1, master_base.shape[1]):
            
            old_flux = master_base[i][j][1:]
            new_wavs = target_wavs
            new_flux = spectres(new_wavs, old_wavs, old_flux, fill=0)
            grid_raw_ages[i, j, 1:] = new_flux

    grid_raw_ages[0].T[0, 1:] = grid_raw_ages[0].T[0, 1:] / 1e9

    return grid_raw_ages

def resample_igm_grid(wavs):
    # Import IGM grid
    raw_igm_grid = fits.open(os.path.dirname(os.path.realpath(__file__)) + '/grids/d_igm_grid_inoue14.fits')[1].data
    igm_wavelengths = np.arange(1.0, 1225.01, 1.0)

    # Resample in wavelength
    grid = np.zeros((raw_igm_grid.shape[0], wavs.shape[0]))

    for i in range(raw_igm_grid.shape[0]):
        
        old_wavs = igm_wavelengths
        old_trans = raw_igm_grid[i]
        new_wavs = wavs
        new_trans = spectres(new_wavs, old_wavs, old_trans)
        grid[i] = new_trans
    
    return grid


hdul = fits.open(os.path.dirname(os.path.realpath(__file__)) + '/grids/ssps.fits')
master_base = np.array((hdul[1].data, hdul[2].data, hdul[3].data, hdul[4].data, hdul[5].data, hdul[6].data, hdul[7].data))

rest_target_wavs = np.arange(100., 50000., 25.)
grid = resample_ssp_wavs(rest_target_wavs)
igm_grid = resample_igm_grid(rest_target_wavs)


'''
# !!AGE RESAMPLING NOT WORKING!!

def resample_ssp_ages(grid_raw_ages, target_ages):

    grid = np.zeros((grid_raw_ages.shape[0], target_ages.shape[0]+1, grid_raw_ages.shape[2]))
    grid[:, 0, :] = grid_raw_ages[0, 0, :]
    grid[:, 1:, 0] = target_ages

    old_ages = grid_raw_ages[0].T[0, 1:] / 1e9

    for i in range(grid_raw_ages.shape[0]):
        for j in range(1, grid_raw_ages.shape[2]):

            old_age_flux = grid_raw_ages[i].T[j, 1:]
            new_ages = target_ages
            new_age_flux = spectres(new_ages, old_ages, old_age_flux, fill=0)
            grid[i].T[j, 1:] = new_age_flux

    return grid

num_agebins = 100
target_ages = np.linspace(0, Planck13.age(0).value, num_agebins)
grid = resample_ssp_ages(grid_resampled, target_ages)
'''



class modelgalaxy(object):
    

    def __init__(self, fit_instructions):
        self.fit_instructions = fit_instructions
        self.universe_age = Planck13.age(0).value
        self.lookback_time = Planck13.lookback_time(self.fit_instructions["redshift"]).value

    
    def sfr(self, age, ageform, tau, massformed):
        
        # Create exp time
        t = ageform - age
        sfr = np.exp(-t/ tau)
        # Enforce no SF before ageform
        sfr[t<0]=0
        # Normalise to massformed
        tdiff = np.diff(age)
        area = np.sum(tdiff * sfr[:-1])
        if area!=0:
            sfr = sfr * ((10**massformed) / area)
        else:
            sfr  = sfr * 0

        if ageform + self.lookback_time > self.universe_age:
            sfr *= 0
            
        return sfr


    def calzetti_dust(self, wavs):
        """ Calculate the ratio A(lambda)/A(V) for the Calzetti et al.
        (2000) attenuation curve. """

        A_lambda = np.zeros_like(wavs)

        wavs_mic = wavs*10**-4

        mask1 = (wavs < 1200.)
        mask2 = (wavs < 6300.) & (wavs >= 1200.)
        mask3 = (wavs < 31000.) & (wavs >= 6300.)

        A_lambda[mask1] = ((wavs_mic[mask1]/0.12)**-0.77
                           * (4.05 + 2.695*(- 2.156 + 1.509/0.12
                                            - 0.198/0.12**2 + 0.011/0.12**3)))

        A_lambda[mask2] = (4.05 + 2.695*(- 2.156
                                         + 1.509/wavs_mic[mask2]
                                         - 0.198/wavs_mic[mask2]**2
                                         + 0.011/wavs_mic[mask2]**3))

        A_lambda[mask3] = 2.659*(-1.857 + 1.040/wavs_mic[mask3]) + 4.05

        A_lambda /= 4.05

        return A_lambda
    

    def transmission_neutral(self, wavs):
        trans = 10**(-self.fit_instructions["dust"]["Av"]*self.calzetti_dust(wavs) / 2.5)
        return trans
    
    def transmission_igm(self, redshift):

        max_redshift=10.
        igm_redshifts = np.arange(0.0, max_redshift + 0.01, 0.01)

        """ Get the IGM transmission at a given redshift. """

        redshift_mask = (igm_redshifts < redshift)
        zred_ind = igm_redshifts[redshift_mask].shape[0]

        zred_fact = ((redshift - igm_redshifts[zred_ind-1])
                     / (igm_redshifts[zred_ind]
                        - igm_redshifts[zred_ind-1]))

        if zred_ind == 0:
            zred_ind += 1
            zred_fact = 0.

        weights = np.array([[1. - zred_fact, zred_fact]])

        igm_trans = np.sum(weights.T*igm_grid[zred_ind-1:zred_ind+1], axis=0)

        igm_trans[np.isnan(igm_trans)] = 1.

        return igm_trans


    def redshift_effects(self, wavs, lum):

        obswavs = (1+self.fit_instructions["redshift"]) * wavs
        lum_distance = Planck13.luminosity_distance(self.fit_instructions["redshift"]).value
        lum = 3.826e33 * lum # change luminosity units to erg/s
        lum_distance = lum_distance * 1e6 * 3.086e16 * 100 # change distance units to cm
        flux = lum / (4 * np.pi * (1+self.fit_instructions["redshift"]) * (lum_distance**2)) # * IGM transmission function
        return obswavs, flux


    def get_spectrum(self):

        # Get model spectra, use a single population
        ssp_age = grid[0].T[0, 1:]
        wavs = grid[0, 0, 1:]
        fluxes = grid[4, 1:, 1:]

        
        agebinwidths = np.diff(ssp_age)
        sfrs = self.sfr(ssp_age[1:], self.fit_instructions["exponential"]["age"], self.fit_instructions["exponential"]["tau"], self.fit_instructions["exponential"]["massformed"])
        ssps = fluxes[1:]
        trans_neutral = self.transmission_neutral(wavs)
        trans_igm = self.transmission_igm(self.fit_instructions["redshift"])

        # Find the total product
        lums = (np.array(([(agebinwidths * sfrs)])).T * ssps) * trans_neutral
        lumtot = np.sum(lums, axis=0)

        lumtot *= trans_igm

        # Add in the effects of ISM dust attenuation (calzetti)
        obswavs, obsflux = self.redshift_effects(wavs, lumtot)

        self.wavs = obswavs
        self.flux = obsflux

        return obswavs, obsflux
    

    def get_photometry(self, wavs, flux, filt_list):
        filt = filter_set(filt_list)
        filt.resample_filter_curves(wavs)
        effwavs = filt.eff_wavs
        phot = filt.get_photometry(flux, 0)
        return effwavs, phot
    

    def plot_spec(self, ax, kwargs):

        #effwavs, phot = self.get_photometry(self.wavs, self.flux)

        self.get_spectrum()

        ax.plot(self.wavs, self.flux, **kwargs)
        ax.set_ylabel('Observed Flux / erg s$^{-1}$ cm$^{-2}$ Å$^{-1}$')
        ax.set_xlabel('Observed Wavelength / Å')
    

    
    def plot_sfh(self, ax):
        ssp_age = grid[0].T[0, 1:]        
        sfrs = self.sfr(ssp_age, self.fit_instructions["exponential"]["age"], self.fit_instructions["exponential"]["tau"], self.fit_instructions["exponential"]["massformed"])

        ax.set_xlabel('Age / Gyr')
        ax.set_ylabel(' Star Formation / M$_{sun}$ Gyr$^{-1}$')
        ax.plot((self.universe_age-self.lookback_time) - ssp_age, sfrs/1e9, color='black', lw=1)
        ax.axvline((self.universe_age-self.lookback_time) - ssp_age[0], 0, 1, color='black', ls='dashed', lw=1)

        #ax.plot(-ssp_age, sfrs/1e9, color='black')
        ax.invert_xaxis()
        ax.set_xlim(self.universe_age,0)
        ax.set_ylim(0, )
    

















import numpy as np
from astropy.io import fits
import os

max_redshift = 10.
igm_redshifts = np.arange(0.0, max_redshift + 0.01, 0.01)
igm_wavelengths = np.arange(1.0, 1225.01, 1.0)
print(igm_redshifts.shape)

coefs = np.loadtxt("/Users/struanstevenson/Desktop/research/CHANTER/lyman_series_coefs_inoue_2014_table2.txt")


def get_Inoue14_trans(rest_wavs, z_obs):
    """ Calculate IGM transmission using Inoue et al. (2014) model. """

    if isinstance(rest_wavs, float):
        rest_wavs = np.array([rest_wavs])

    tau_LAF_LS = np.zeros((39, rest_wavs.shape[0]))
    tau_DLA_LS = np.zeros((39, rest_wavs.shape[0]))
    tau_LAF_LC = np.zeros(rest_wavs.shape[0])
    tau_DLA_LC = np.zeros(rest_wavs.shape[0])

    # Populate tau_LAF_LS
    for j in range(39):

        if z_obs < 1.2:
            wav_slice = ((rest_wavs*(1.+z_obs) > coefs[j, 1])
                         & (rest_wavs*(1.+z_obs)
                            < (1+z_obs)*coefs[j, 1]))

            tau_LAF_LS[j, wav_slice] = (coefs[j, 2]
                                        * (rest_wavs[wav_slice]
                                           * (1.+z_obs)/coefs[j, 1])**1.2)

        elif z_obs < 4.7:
            wav_slice_1 = ((rest_wavs*(1.+z_obs) > coefs[j, 1])
                           & (rest_wavs*(1.+z_obs) < 2.2*coefs[j, 1]))
            wav_slice_2 = ((rest_wavs*(1.+z_obs) > 2.2*coefs[j, 1])
                           & (rest_wavs*(1.+z_obs)
                              < (1+z_obs)*coefs[j, 1]))

            tau_LAF_LS[j, wav_slice_1] = (coefs[j, 2]
                                          * (rest_wavs[wav_slice_1]
                                             * (1.+z_obs)/coefs[j, 1])**1.2)

            tau_LAF_LS[j, wav_slice_2] = (coefs[j, 3]
                                          * (rest_wavs[wav_slice_2]
                                             * (1.+z_obs)/coefs[j, 1])**3.7)

        else:
            wav_slice_1 = ((rest_wavs*(1.+z_obs) > coefs[j, 1])
                           & (rest_wavs*(1.+z_obs) < 2.2*coefs[j, 1]))

            wav_slice_2 = ((rest_wavs*(1.+z_obs) > 2.2*coefs[j, 1])
                           & (rest_wavs*(1.+z_obs) < 5.7*coefs[j, 1]))

            wav_slice_3 = ((rest_wavs*(1.+z_obs) > 5.7*coefs[j, 1])
                           & (rest_wavs*(1.+z_obs)
                              < (1+z_obs)*coefs[j, 1]))

            tau_LAF_LS[j, wav_slice_1] = (coefs[j, 2]
                                          * (rest_wavs[wav_slice_1]
                                             * (1.+z_obs)/coefs[j, 1])**1.2)

            tau_LAF_LS[j, wav_slice_2] = (coefs[j, 3]
                                          * (rest_wavs[wav_slice_2]
                                             * (1.+z_obs)/coefs[j, 1])**3.7)

            tau_LAF_LS[j, wav_slice_3] = (coefs[j, 4]
                                          * (rest_wavs[wav_slice_3]
                                             * (1.+z_obs)/coefs[j, 1])**5.5)

    # Populate tau_DLA_LS
    for j in range(39):

        if z_obs < 2.0:
            wav_slice = ((rest_wavs*(1.+z_obs) > coefs[j, 1])
                         & (rest_wavs*(1.+z_obs)
                            < (1+z_obs)*coefs[j, 1]))

            tau_DLA_LS[j, wav_slice] = (coefs[j, 5]
                                        * (rest_wavs[wav_slice]
                                           * (1.+z_obs)/coefs[j, 1])**2.0)

        else:
            wav_slice_1 = ((rest_wavs*(1.+z_obs) > coefs[j, 1])
                           & (rest_wavs*(1.+z_obs) < 3.0*coefs[j, 1]))

            wav_slice_2 = ((rest_wavs*(1.+z_obs) > 3.0*coefs[j, 1])
                           & (rest_wavs*(1.+z_obs) < (1+z_obs)
                              * coefs[j, 1]))

            tau_DLA_LS[j, wav_slice_1] = (coefs[j, 5]
                                          * (rest_wavs[wav_slice_1]
                                             * (1.+z_obs)/coefs[j, 1])**2.0)

            tau_DLA_LS[j, wav_slice_2] = (coefs[j, 6]
                                          * (rest_wavs[wav_slice_2]
                                             * (1.+z_obs)/coefs[j, 1])**3.0)

    # Populate tau_LAF_LC
    if z_obs < 1.2:
        wav_slice = ((rest_wavs*(1.+z_obs) > 911.8)
                     & (rest_wavs*(1.+z_obs) < 911.8*(1.+z_obs)))

        tau_LAF_LC[wav_slice] = (0.325*((rest_wavs[wav_slice]
                                         * (1.+z_obs)/911.8)**1.2
                                        - (((1+z_obs)**-0.9)
                                           * (rest_wavs[wav_slice]
                                           * (1.+z_obs)/911.8)**2.1)))

    elif z_obs < 4.7:
        wav_slice_1 = ((rest_wavs*(1.+z_obs) > 911.8)
                       & (rest_wavs*(1.+z_obs) < 911.8*2.2))

        wav_slice_2 = ((rest_wavs*(1.+z_obs) > 911.8*2.2)
                       & (rest_wavs*(1.+z_obs) < 911.8*(1.+z_obs)))

        tau_LAF_LC[wav_slice_1] = (((2.55*10**-2)*((1+z_obs)**1.6)
                                    * (rest_wavs[wav_slice_1]
                                    * (1.+z_obs)/911.8)**2.1)
                                   + (0.325*((rest_wavs[wav_slice_1]
                                      * (1.+z_obs)/911.8)**1.2))
                                   - (0.25*((rest_wavs[wav_slice_1]
                                             * (1.+z_obs)/911.8)**2.1)))

        tau_LAF_LC[wav_slice_2] = ((2.55*10**-2)
                                   * (((1.+z_obs)**1.6)
                                      * ((rest_wavs[wav_slice_2]
                                          * (1.+z_obs)/911.8)**2.1)
                                      - ((rest_wavs[wav_slice_2]
                                          * (1.+z_obs)/911.8)**3.7)))

    else:
        wav_slice_1 = ((rest_wavs*(1.+z_obs) > 911.8)
                       & (rest_wavs*(1.+z_obs) < 911.8*2.2))

        wav_slice_2 = ((rest_wavs*(1.+z_obs) > 911.8*2.2)
                       & (rest_wavs*(1.+z_obs) < 911.8*5.7))

        wav_slice_3 = ((rest_wavs*(1.+z_obs) > 911.8*5.7)
                       & (rest_wavs*(1.+z_obs) < 911.8*(1.+z_obs)))

        tau_LAF_LC[wav_slice_1] = (((5.22*10**-4)*((1+z_obs)**3.4)
                                    * (rest_wavs[wav_slice_1]
                                       * (1.+z_obs)/911.8)**2.1)
                                   + (0.325*(rest_wavs[wav_slice_1]
                                      * (1.+z_obs)/911.8)**1.2)
                                   - ((3.14*10**-2)*((rest_wavs[wav_slice_1]
                                      * (1.+z_obs)/911.8)**2.1)))

        tau_LAF_LC[wav_slice_2] = (((5.22*10**-4)*((1+z_obs)**3.4)
                                    * (rest_wavs[wav_slice_2]
                                       * (1.+z_obs)/911.8)**2.1)
                                   + (0.218*((rest_wavs[wav_slice_2]
                                             * (1.+z_obs)/911.8)**2.1))
                                   - ((2.55*10**-2)*((rest_wavs[wav_slice_2]
                                                      * (1.+z_obs)
                                                      / 911.8)**3.7)))

        tau_LAF_LC[wav_slice_3] = ((5.22*10**-4)
                                   * (((1+z_obs)**3.4)
                                      * (rest_wavs[wav_slice_3]
                                         * (1.+z_obs)/911.8)**2.1
                                      - (rest_wavs[wav_slice_3]
                                         * (1.+z_obs)/911.8)**5.5))

    # Populate tau_DLA_LC
    if z_obs < 2.0:
        wav_slice = ((rest_wavs*(1.+z_obs) > 911.8)
                     & (rest_wavs*(1.+z_obs) < 911.8*(1.+z_obs)))

        tau_DLA_LC[wav_slice] = (0.211*((1+z_obs)**2.)
                                 - (7.66*10**-2)*(((1+z_obs)**2.3)
                                                  * (rest_wavs[wav_slice]
                                                     * (1.+z_obs)/911.8)**-0.3)
                                 - 0.135*((rest_wavs[wav_slice]
                                           * (1.+z_obs)/911.8)**2.0))

    else:
        wav_slice_1 = ((rest_wavs*(1.+z_obs) > 911.8)
                       & (rest_wavs*(1.+z_obs) < 911.8*3.0))

        wav_slice_2 = ((rest_wavs*(1.+z_obs) > 911.8*3.0)
                       & (rest_wavs*(1.+z_obs) < 911.8*(1.+z_obs)))

        tau_DLA_LC[wav_slice_1] = (0.634 + (4.7*10**-2)*(1.+z_obs)**3.
                                   - ((1.78*10**-2)*((1.+z_obs)**3.3)
                                      * (rest_wavs[wav_slice_1]
                                         * (1.+z_obs)/911.8)**-0.3)
                                   - (0.135*(rest_wavs[wav_slice_1]
                                             * (1.+z_obs)/911.8)**2.0)
                                   - 0.291*(rest_wavs[wav_slice_1]
                                            * (1.+z_obs)/911.8)**-0.3)

        tau_DLA_LC[wav_slice_2] = ((4.7*10**-2)*(1.+z_obs)**3.
                                   - ((1.78*10**-2)*((1.+z_obs)**3.3)
                                      * (rest_wavs[wav_slice_2]
                                         * (1.+z_obs)/911.8)**-0.3)
                                   - ((2.92*10**-2)
                                      * (rest_wavs[wav_slice_2]
                                         * (1.+z_obs)/911.8)**3.0))

    tau_LAF_LS_sum = np.sum(tau_LAF_LS, axis=0)
    tau_DLA_LS_sum = np.sum(tau_DLA_LS, axis=0)

    tau = tau_LAF_LS_sum + tau_DLA_LS_sum + tau_LAF_LC + tau_DLA_LC

    return np.exp(-tau)


def make_table(z_array, rest_wavs):
    """ Make up the igm absorption table used by bagpipes. """

    print("BAGPIPES: Generating IGM absorption table.")

    d_IGM_grid = np.zeros((z_array.shape[0], rest_wavs.shape[0]))

    for i in range(z_array.shape[0]):
        d_IGM_grid[i, :] = get_Inoue14_trans(rest_wavs, z_array[i])

    hdulist = fits.HDUList(hdus=[fits.PrimaryHDU(),
                                 fits.ImageHDU(name="trans", data=d_IGM_grid),
                                 fits.ImageHDU(name="wavs", data=rest_wavs),
                                 fits.ImageHDU(name="zred", data=z_array)])

    if os.path.exists("d_igm_grid_inoue14.fits"):
        os.system("rm " + "d_igm_grid_inoue14.fits")

    hdulist.writeto("d_igm_grid_inoue14.fits")


make_table(igm_redshifts, igm_wavelengths)


import setuptools
from setuptools import setup

setup(
    name='chanter',

    version='0.0.4',

    description='Galaxy spectral fitting',

    author='Struan Stevenson',

    author_email='struan.stevenson@ed.ac.uk',

    packages= setuptools.find_packages(),

    package_data = {'': ['*.txt', '*.fits'],},  

    install_requires=["numpy", "astropy", "matplotlib", "spectres", "nautilus-sampler"],

)